import math
from .types import *
from .constants import *
from . import validate


def is_ccw(linear_ring: Points):
    validate.linear_ring(linear_ring)
    area = 0.0
    if len(linear_ring) > 2:
        if (
            linear_ring[0][0] == linear_ring[-1][0]
            and linear_ring[0][1] == linear_ring[-1][1]
        ):
            length = len(linear_ring) - 1
        else:
            length = len(linear_ring)

        for i in range(length - 1):
            area = (
                area
                + linear_ring[i][0] * linear_ring[i + 1][1]
                - linear_ring[i][1] * linear_ring[i + 1][0]
            )

        area = (
            area
            + linear_ring[length - 1][0] * linear_ring[0][1]
            - linear_ring[length - 1][1] * linear_ring[0][0]
        )

    return area >= 0


def within(point: Point, linear_ring: Points, include_border: bool = False):
    """
    Winding Number Algorithm
    """
    validate.points([point])
    validate.linear_ring(linear_ring)
    [x, y] = point
    theta: float = 0
    for i in range(len(linear_ring)):
        x1, y1 = linear_ring[i - 1]
        x2, y2 = linear_ring[i]
        x1 -= x
        y1 -= y
        x2 -= x
        y2 -= y

        cv = x1 * x2 + y1 * y2
        sv = x1 * y2 - x2 * y1
        if abs(sv) < EPSILON and cv <= 0:
            return include_border
        theta += math.atan2(sv, cv)
    return abs(theta) > 1


def intersection(p1: Point, p2: Point, p3: Point, p4: Point):
    """
    When p1 -> p2, p3 -> p4 are crossing or points more than 3 are on a line,
    return True
    """
    validate.points([p1, p2, p3, p4])
    if p1[0] >= p2[0]:
        if (p1[0] < p3[0] and p1[0] < p4[0]) or (p2[0] > p3[0] and p2[0] > p4[0]):
            return False
    else:
        if (p2[0] < p3[0] and p2[0] < p4[0]) or (p1[0] > p3[0] and p1[0] > p4[0]):
            return False

    if p1[1] >= p2[1]:
        if (p1[1] < p3[1] and p1[1] < p4[1]) or (p2[1] > p3[1] and p2[1] > p4[1]):
            return False
    else:
        if (p2[1] < p3[1] and p2[1] < p4[1]) or (p1[1] > p3[1] and p1[1] > p4[1]):
            return False

    if ((p1[0] - p2[0]) * (p3[1] - p1[1]) + (p1[1] - p2[1]) * (p1[0] - p3[0])) * (
        (p1[0] - p2[0]) * (p4[1] - p1[1]) + (p1[1] - p2[1]) * (p1[0] - p4[0])
    ) > 0:
        return False
    if ((p3[0] - p4[0]) * (p1[1] - p3[1]) + (p3[1] - p4[1]) * (p3[0] - p1[0])) * (
        (p3[0] - p4[0]) * (p2[1] - p3[1]) + (p3[1] - p4[1]) * (p3[0] - p2[0])
    ) > 0:
        return False

    return True


def selfintersection(linear_ring: Points):
    """
    not support warp polygon.
    """
    validate.linear_ring(linear_ring)
    if len(linear_ring) == 4:
        return (
            abs(
                linear_ring[0][1] * (linear_ring[1][0] - linear_ring[2][0])
                + linear_ring[1][1] * (linear_ring[2][0] - linear_ring[0][0])
                + linear_ring[2][1] * (linear_ring[0][0] - linear_ring[1][0])
            )
            < EPSILON
        )

    lines = [[linear_ring[i], linear_ring[i + 1]] for i in range(len(linear_ring) - 2)]

    def check(lines, start=0):
        if start + 2 >= len(lines):
            return False

        l1 = lines[start]
        for i in range(start + 2, len(lines)):
            l2 = lines[i]
            if intersection(*l1, *l2):
                return True
        return check(lines, start + 1)

    return check(lines)


from dataclasses import dataclass
from .types import *
from .calc import *
from .utils import *
from .exceptions import *
from . import validate


def _warp_within(lon: float):
    while lon < -180:
        lon = lon + 360
    while 180 < lon:
        lon = lon - 360
    return lon


def _is_crossing_antimeridian(lon1: float, lon2: float):
    """
    The distance in the longitude between the points must be less than 180deg.

    lon1=-10, lon2=10 -> False
    lon1=170, lon2=190 -> True
    lon1=170, lon2=-170 -> True
    """
    if abs(lon1 - lon2) > 360:
        return True
    lon1 = _warp_within(lon1)
    lon2 = _warp_within(lon2)

    if lon1 * lon2 > 0:
        return False

    return abs(lon1 - lon2) > 180


def _crossing_antimeridian_point_lat(p1: Point, p2: Point):
    """
    p1=[-150, 0], p2=[170, 20]) -> 15
    p1=[-20,-10], p2=[10, 170] -> 85
    p1=[-20, 66], p2=[10, 0]) -> 34
    p1=[-160, 32], p2=[170, 20]) -> 24
    """
    [x1, y1] = p1
    [x2, y2] = p2
    if x1 < 0:
        x1 = _warp_within(x1) + 360
    if x2 < 0:
        x2 = _warp_within(x2) + 360
    return linear_y([x1, y1], [x2, y2], 180)


@dataclass
class Ring:
    within: list
    overflow: list


def cut_ring_at_antimeridian(
    linear_ring: Points, overflowing: bool = False, allow_selfintersection: bool = False
):
    """
    When overflowing is True, points are right side of 180 degrees.
    The overflowing flag is used when recursive.
    The distance in the longitude between the points must be less than 180deg.
    """
    validate.linear_ring(linear_ring)

    crossing_lats = []

    for i in range(len(linear_ring) - 1):
        if _is_crossing_antimeridian(linear_ring[i][0], linear_ring[i + 1][0]):
            crossing_lats.append(
                {
                    "from": i,
                    "to": i + 1,
                    "lat": _crossing_antimeridian_point_lat(
                        linear_ring[i], linear_ring[i + 1]
                    ),
                }
            )
    if len(crossing_lats) < 2:
        return (
            Ring(within=[linear_ring], overflow=[])
            if not overflowing
            else Ring(within=[], overflow=[linear_ring])
        )

    sorted(crossing_lats, key=lambda p: p["lat"])
    start = crossing_lats[-1]
    end = crossing_lats[-2]

    ret: Ring = Ring(within=[], overflow=[])

    def _cutting(start, end):
        ring_index = start["to"] if not start["to"] == len(linear_ring) - 1 else 0
        bound_lon = 180 if linear_ring[ring_index][0] >= 0 else -180

        rtn = {"overflowing": bound_lon < 0, "linear_ring": []}
        rtn["linear_ring"].append(
            [
                bound_lon,
                _crossing_antimeridian_point_lat(
                    linear_ring[start["from"]], linear_ring[start["to"]]
                ),
            ]
        )
        rtn["linear_ring"].append(linear_ring[start["to"]])

        while ring_index != end["from"]:
            ring_index = (
                ring_index + 1 if not ring_index + 1 == len(linear_ring) - 1 else 0
            )
            rtn["linear_ring"].append(linear_ring[ring_index])
        rtn["linear_ring"].append(
            [
                bound_lon,
                _crossing_antimeridian_point_lat(
                    linear_ring[end["from"]], linear_ring[end["to"]]
                ),
            ]
        )
        rtn["linear_ring"].append(rtn["linear_ring"][0])
        return rtn

    cut1 = _cutting(start, end)
    result1 = cut_ring_at_antimeridian(
        cut1["linear_ring"],
        cut1["overflowing"],
        allow_selfintersection=allow_selfintersection,
    )
    ret.within.extend(result1.within)
    ret.overflow.extend(result1.overflow)

    cut2 = _cutting(end, start)
    result2 = cut_ring_at_antimeridian(
        cut2["linear_ring"],
        cut2["overflowing"],
        allow_selfintersection=allow_selfintersection,
    )
    ret.within.extend(result2.within)
    ret.overflow.extend(result2.overflow)

    if not allow_selfintersection:
        for ring in ret.within:
            if selfintersection(ring):
                raise InvalidSelfintersection

        for ring in ret.overflow:
            if selfintersection(ring):
                raise InvalidSelfintersection

    return ret


from .types import *
from .exceptions import *


def points(points: Points):
    for p in points:
        if not len(p) == 2:
            raise InvalidPoint


def linear_ring(linear_ring: Points):
    points(linear_ring)
    if (
        len(linear_ring) <= 3
        or not linear_ring[0][0] == linear_ring[-1][0]
        or not linear_ring[0][1] == linear_ring[-1][1]
    ):
        raise InvalidLinearRing


class InvalidPoint(Exception):
    def __str__(self):
        return "point's dimension must be 2."


class InvalidLinearRing(Exception):
    """invalid linear ring."""


class EnclosingBothPoles(Exception):
    """not support linear ring enclosing north and south poles."""


class InvalidLinearRingEnclosingPole(Exception):
    """invalid linear ring enclosing the pole."""


class InvalidBounds(Exception):
    """invalid bounds."""


class NotAllowedWarpBounds(Exception):
    """not support warpping bounds."""


class NotAllowedCwLinearRing(Exception):
    """not support cw linear ring."""


class InvalidSelfintersection(Exception):
    """invalid selfintersection."""


class FalidCuttingAntimeridian(Exception):
    """falid cutting antimeridian."""


import sys
from functools import reduce
from typing import Sequence
from pyproj import CRS, Transformer
from .types import *
from .exceptions import *
from .constants import *
from .utils import *
from .calc import *
from .lonlat import *
from . import validate


def _transform(points: Points, src_crs: CRS, dst_crs: CRS):
    if src_crs == dst_crs:
        return points
    transformer = Transformer.from_crs(src_crs, dst_crs, always_xy=True)

    return [list(transformer.transform(*p)) for p in points]


def _transform_enclosing_pole_ring(
    linear_ring: Points, src_crs: CRS, partition: int, north: bool = True
):
    length = len(linear_ring) - 1
    temp_CRS = (
        ARCTIC_POLAR_STEROGRAPHIC_CRS if north else ANTARCTIC_POLAR_STEROGRAPHIC_CRS
    )

    end_y = sys.maxsize if north else -1 * sys.maxsize
    pole_lat = 90 if north else -90
    polar_sterographic_linear_ring = _transform(linear_ring, src_crs, temp_CRS)

    crossing_ys = []
    for i in range(length):
        if intersection(
            polar_sterographic_linear_ring[i],
            polar_sterographic_linear_ring[i + 1],
            [0, 0],
            [0, end_y],
        ):
            crossing_ys.append(
                {
                    "from": i,
                    "to": i + 1,
                    "y": linear_y(
                        polar_sterographic_linear_ring[i],
                        polar_sterographic_linear_ring[i + 1],
                        0,
                    ),
                }
            )
    if len(crossing_ys) == 0:
        raise InvalidLinearRingEnclosingPole
    sorted(crossing_ys, key=lambda p: p["y"])

    crossing_y = crossing_ys[0] if north else crossing_ys[-1]

    ret = []
    for i in range(length):
        if i == crossing_y["from"]:
            transformed = _transform(
                linear_interpolate(
                    polar_sterographic_linear_ring[crossing_y["from"]],
                    [0, crossing_y["y"]],
                    partition,
                ),
                temp_CRS,
                EPSG4326_CRS,
            )

            ret.extend(transformed[:-1])
            if transformed[0][0] >= 0:
                ret.extend(
                    [
                        [180, transformed[-1][1]],
                        [180, pole_lat],
                        [-180, pole_lat],
                        [-180, transformed[-1][1]],
                    ]
                )
            else:
                ret.extend(
                    [
                        [-180, transformed[-1][1]],
                        [-180, pole_lat],
                        [180, pole_lat],
                        [180, transformed[-1][1]],
                    ]
                )
            ret.extend(
                _transform(
                    linear_interpolate(
                        [0, crossing_y["y"]],
                        polar_sterographic_linear_ring[crossing_y["to"]],
                        partition,
                    ),
                    temp_CRS,
                    EPSG4326_CRS,
                )[1:-1]
            )
        else:
            ret.extend(
                _transform(
                    linear_interpolate(
                        polar_sterographic_linear_ring[i],
                        polar_sterographic_linear_ring[i + 1],
                        partition,
                    ),
                    temp_CRS,
                    EPSG4326_CRS,
                )[:-1]
            )
    ret.append(ret[0])

    return ret


def transform_ring(linear_ring: Points, src_crs: CRS, partition: int = 0):
    """
    not support linear rings of including both poles.
    """
    validate.linear_ring(linear_ring)
    length = len(linear_ring) - 1

    north_pole = _transform([[0, 90]], EPSG4326_CRS, src_crs)[0]
    enclosing_north_pole = within(north_pole, linear_ring)
    south_pole = _transform([[0, -90]], EPSG4326_CRS, src_crs)[0]
    enclosing_south_pole = within(south_pole, linear_ring)
    if enclosing_north_pole and enclosing_south_pole:
        raise EnclosingBothPoles

    if enclosing_north_pole:
        return _transform_enclosing_pole_ring(
            linear_ring, src_crs, partition=partition, north=True
        )
    if enclosing_south_pole:
        return _transform_enclosing_pole_ring(
            linear_ring, src_crs, partition=partition, north=False
        )

    interpolated_linear_ring = []
    for i in range(length):
        interpolated_linear_ring.extend(
            linear_interpolate(linear_ring[i], linear_ring[i + 1], partition)[:-1]
        )
    interpolated_linear_ring.append(interpolated_linear_ring[0])

    if src_crs == EPSG4326_CRS:
        return interpolated_linear_ring

    return _transform(interpolated_linear_ring, src_crs, EPSG4326_CRS)


def transform_bbox(
    src_bbox: Sequence[float], src_crs: CRS, partition: int = 9, expand: bool = False
):
    """
    input bbox is not allowed warp.
    """
    if not (len(src_bbox) == 4 or len(src_bbox) == 6):
        raise InvalidBounds

    has_height = len(src_bbox) == 6
    left = src_bbox[0]
    bottom = src_bbox[1]
    right = src_bbox[3] if has_height else src_bbox[2]
    top = src_bbox[4] if has_height else src_bbox[3]

    if right < left:
        raise NotAllowedWarpBounds

    points = transform_ring(
        [[left, bottom], [right, bottom], [right, top], [left, top], [left, bottom]],
        src_crs,
        partition=partition,
    )

    ys = [p[1] for p in points]
    if is_ccw(points):
        xs = [p[0] for p in points]
        if len(src_bbox) == 6:
            return [min(xs), min(ys), src_bbox[2], max(xs), max(ys), src_bbox[5]]
        return [min(xs), min(ys), max(xs), max(ys)]

    try:
        bounds_ring = cut_ring_at_antimeridian(points)
        xs1 = [p[0] for p in reduce(lambda a, b: a + b, bounds_ring.within)]
        xs2 = [p[0] for p in reduce(lambda a, b: a + b, bounds_ring.overflow)]
        if len(src_bbox) == 6:
            return [
                min(xs1),
                min(ys),
                src_bbox[2],
                max(xs2) + 360 if expand else max(xs2),
                max(ys),
                src_bbox[5],
            ]
        return [min(xs1), min(ys), max(xs2) + 360 if expand else max(xs2), max(ys)]
    except:
        raise FalidCuttingAntimeridian


def geojson_from_linear_ring(linear_ring: Points, src_crs: CRS, partition: int = 9):
    if not is_ccw(linear_ring):
        raise NotAllowedCwLinearRing
    points = transform_ring(linear_ring, src_crs, partition=partition)
    ys = [p[1] for p in points]

    if is_ccw(points):
        xs = [p[0] for p in points]
        return {
            "type": "Feature",
            "bbox": [min(xs), min(ys), max(xs), max(ys)],
            "properties": {},
            "geometry": {"type": "Polygon", "coordinates": [points]},
        }

    try:
        ring = cut_ring_at_antimeridian(points)
        xs1 = [p[0] for p in reduce(lambda a, b: a + b, ring.within)]
        xs2 = [p[0] for p in reduce(lambda a, b: a + b, ring.overflow)]
        coordinates = []
        coordinates.extend([[r] for r in ring.within])
        coordinates.extend([[r] for r in ring.overflow])
        return {
            "type": "Feature",
            "bbox": [min(xs1), min(ys), max(xs2), max(ys)],
            "properties": {},
            "geometry": {"type": "MultiPolygon", "coordinates": coordinates},
        }

    except:
        raise FalidCuttingAntimeridian


def geojson_from_corner_coordinates(
    upper_left: Point,
    lower_left: Point,
    upper_right: Point,
    lower_right: Point,
    src_crs: CRS,
    partition: int = 9,
):
    return geojson_from_linear_ring(
        [upper_left, lower_left, lower_right, upper_right, upper_left],
        src_crs,
        partition=partition,
    )



import os
from setuptools import setup, find_packages


def load_readme():
    with open("README.md", encoding="utf-8") as f:
        return f.read()


def read_requirements():
    """Parse requirements from requirements.txt."""
    reqs_path = os.path.join(".", "requirements.txt")
    with open(reqs_path, "r") as f:
        requirements = [line.rstrip() for line in f]
    return requirements


version = None
with open("to4326/__init__.py", "r") as fp:
    for line in fp:
        if line.startswith("__version__"):
            version = line.split("=")[1].strip().strip("\"'")
            break

setup(
    name="to4326",
    version=str(version),
    long_description_content_type="text/markdown",
    long_description=load_readme(),
    packages=find_packages(),
    author="yonda",
    author_email="yonda.fountain@gmail.com",
    url="https://github.com/yonda-yonda/to4326",
    install_requires=read_requirements(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    python_requires=">=3.7",
)


from .types import *
from . import validate


def linear_x(p1: Point, p2: Point, y: float):
    validate.points([p1, p2])
    if p2[1] == p1[1]:
        return p1[0]
    return p1[0] + (y - p1[1]) * (p2[0] - p1[0]) / (p2[1] - p1[1])


def linear_y(p1: Point, p2: Point, x: float):
    validate.points([p1, p2])
    if p2[0] == p1[0]:
        return p1[1]
    return p1[1] + (x - p1[0]) * (p2[1] - p1[1]) / (p2[0] - p1[0])


def linear_interpolate(p1: Point, p2: Point, partition: int = 0):
    validate.points([p1, p2])

    if partition <= 0:
        return [p1, p2]
    ret = []

    def calc(i):
        if p2[0] == p1[0]:
            x = p1[0]
            y = p1[1] + i * (p2[1] - p1[1]) / (partition + 1)
        elif p2[1] == p1[1]:
            x = p1[0] + i * (p2[0] - p1[0]) / (partition + 1)
            y = p1[1]
        else:
            x = p1[0] + i * (p2[0] - p1[0]) / (partition + 1)
            y = linear_y(p1, p2, x)

        return [x, y]

    ret = [calc(i) for i in range(partition + 1)]
    ret.append(p2)
    return ret


from pyproj import CRS

EPSG4326_CRS = CRS.from_epsg(4326)
ARCTIC_POLAR_STEROGRAPHIC_CRS = CRS.from_epsg(3995)
ANTARCTIC_POLAR_STEROGRAPHIC_CRS = CRS.from_epsg(3031)
SOUTH_POLE_CRS = CRS.from_epsg(3031)
EPSILON = 1e-8
